module user_output_eddyflux

  use binas,      only : ae, twopi, grav
  use dims,       only : im, jm, lm, dx, dy, xref, yref, xbeg, ybeg, xend, yend
  use dims,       only : nregions, region_name, itaur, xcyc
  use chem_param, only : ntrace, ntracet, fscale
  use file_hdf,   only : THdfFile, Init, Done, TSds, SetDim
  use global_data,only : InputDir

  implicit none

  private
  
  public :: evaluate_eddyflux, read_eddyfluxlist, eddystat, eddyfluxs
  public :: init_eddyflux_output, reset_eddyflux_accumulator
  public :: write_eddyfluxfiles, eddyfluxlistFFilename , output_eddyflux
  public :: close_eddyfluxfiles, write_eddyflux
  public :: n_eddystat, io_nrecords

  ! MODIFIED by Peter Beramaschi.
  ! PARALLEL version Maarten Krol Jul 2003
  ! jun 2005: switch to HDF output file

  character(len=100)           :: eddyfluxlistFFilename 
  character(len=100)           :: eddyfluxOutFFilename 
  logical                      :: output_eddyfluxs_meteo = .true.
  logical                      :: write_eddyfluxfiles = .true.
  integer                      :: cstatus

  integer,save                 :: n_eddystat         ! number of eddyfluxs 
  ! list of species for output
  integer, dimension(2)        :: eddyflux_speclist
  ! number of species for output
  integer,parameter            :: eddyflux_nspec=3  ! fossil fuel and biosphere flux
  character(len=10), dimension(eddyflux_nspec), parameter :: names = (/'ff    ','bio       ','oceans    '/)
  ! max number of meteo data (for output in eddyflux output file)
  integer, parameter           :: n_metdat = 7
  integer, parameter           :: n_interpol = 3  ! number of eddyflux interpolation methods
  character(len=4), dimension(n_interpol), parameter :: c_interpol = (/'_grd','_slp','_int'/)
  character(len=12),dimension(n_metdat  ), parameter :: C_metdat   =  &
               (/'Temperature', &
                 'Pressure   ',&
                 'uwind      ',&
                 'vwind      ',&
                 'windspeed  ',&
                 'winddir    ',&
                 'blh        '/)
  character(len=12),dimension(n_metdat  ), parameter :: U_metdat   =  &
               (/'K           ', &
                 'hPa         ', &
                 'm/s         ', &
                 'm/s         ', &
                 'm/s         ', &
                 'degrees wrtN', &
                 'm           '/)

  type(THdfFile)               :: io_hdf          ! contains HDF output file
  integer                      :: io_record = 0
  integer,save                 :: io_nrecords 
  ! 'open' datasets of the io_hdf output file:
  type(TSds), dimension(:,:), allocatable :: Sds
  type(TSds), dimension(n_metdat)         :: Sds_meteo
  type(TSds)                              :: Sds_idate

  type eddyfluxs
     !cmk character(len=7)           :: ident     ! eddyflux identifier
     character(len=13)           :: ident     ! eddyflux identifier
     character(len=50)          :: name      ! eddyflux name
     real                       :: lat       ! latitude of eddyflux
     real                       :: lon       ! longitude of eddyflux
     real                       :: height    ! height of eddyflux
     integer                    :: index     ! index of eddyflux
     integer                    :: region 
     integer                    :: is
     integer                    :: js
     integer                    :: ls
     real,dimension(n_metdat)   :: metdat               ! meteodata (interpolated) at eddyfluxs
     real,dimension(:,:),pointer :: c_mean              ! spcies, interpolation
     real,dimension(:,:),pointer :: c_std
     integer                     :: ncount              ! for averages
  end type eddyfluxs

  type(eddyfluxs), dimension(:), allocatable  :: eddystat

contains



  subroutine free_eddyfluxsfields
    use Partools, only : myid, root
    implicit none
    integer :: i_stat
    if (myid /= root) return
    do i_stat = 1, n_eddystat
       deallocate(eddystat(i_stat)%c_mean)
       deallocate(eddystat(i_stat)%c_std )
    enddo
    deallocate( eddystat )
  end subroutine free_eddyfluxsfields



  subroutine read_eddyfluxlist
    use toolbox,      only : escape_tm
#ifdef MPI  
    use mpi_const, only: myid, root, mpi_integer, mpi_character, &
         my_real, mpi_comm_world, ierr
#else
    use ParTools, only: myid, root
#endif
    use GO,          only : TrcFile, Init, Done, ReadRc
    use global_data, only : rcfile
    

    implicit none

    ! local
    integer                           :: slen1 !string length
    character(len=200)                :: dummy
    integer                           :: fstatus 
    integer                           :: i_stat
    integer                           :: region
    integer,dimension(6)              :: idate_sim
    integer                           :: itau_sim, isim, ispec
    ! x,y resolution (in degrees) for current region 
    real                              :: dxr, dyr 
    character(len=8)                  :: name_spec    !CMK changed from 6-->8
    integer                           :: sunit0=110, iostat
    type(TrcFile)                     :: rcF

    ! start
    if ( myid == root ) then
          !=========================================
          ! open eddyfluxFile (with list of eddyfluxs)
          !=========================================
          open(UNIT=sunit0, FORM='FORMATTED', &
               FILE=trim(InputDir)//'/'//trim(eddyfluxlistFFilename), &
               STATUS='OLD', ERR=1000)      
          read(sunit0, '(a)') dummy
          read(sunit0, '(a)') dummy

          ! count number of eddyfluxs
          n_eddystat=0
          stat_loop: DO
             read(sunit0,  '(a)', END=100) dummy
             n_eddystat = n_eddystat+1 
          end do stat_loop
100       continue

          print *, '_________________________________________________' &
               //'_______________________________'
          print *, n_eddystat, ' eddyfluxs read from file: ', eddyfluxlistFFilename

          allocate(eddystat(n_eddystat))

          do i_stat = 1, n_eddystat
             allocate(eddystat(i_stat)%c_mean(eddyflux_nspec,n_interpol))
             allocate(eddystat(i_stat)%c_std(eddyflux_nspec,n_interpol))
          enddo
          dyr = dy/yref(1)
          dxr = dx/xref(1)

          !===============
          ! read eddyfluxs
          !===============

          rewind(sunit0)
          read(sunit0, '(a)') dummy
          read(sunit0, '(a)') dummy
          do i_stat=1, n_eddystat 

             read(sunit0, '(a3,2x,a17,f11.6,2x,f7.6)') &
                  eddystat(i_stat)%ident, &
                  eddystat(i_stat)%name, &
                  eddystat(i_stat)%lon, &
                  eddystat(i_stat)%lat

             print '(a3,2x,a17,f11.6,2x,f11.6)', &
                  eddystat(i_stat)%ident, &
                  eddystat(i_stat)%name, &
                  eddystat(i_stat)%lon, &
                  eddystat(i_stat)%lat


             eddystat(i_stat)%index = i_stat

             ! assume global region as default for average mixing ratios
             eddystat(i_stat)%region = 1
             eddystat(i_stat)%height = 20.0
             eddystat(i_stat)%is = &
                  int((eddystat(i_stat)%lon-float(xbeg(1)))/dxr + 0.99999) 
             eddystat(i_stat)%js = &
                  int((eddystat(i_stat)%lat-float(ybeg(1)))/dyr + 0.99999)
          end do

          ! close eddyfluxfile
          close(sunit0)
          do i_stat=1, n_eddystat
            do region=1, nregions
      
              dyr = dy/yref(region)
              dxr = dx/xref(region)

              if ( (eddystat(i_stat)%lon .ge. xbeg(region) .and. &
                    eddystat(i_stat)%lon .le. xend(region)) .and.      &
                   (eddystat(i_stat)%lat .ge. ybeg(region) .and. &
                    eddystat(i_stat)%lat .le. yend(region) ) ) then
                   !=====================
                   ! eddyflux is in region
                   !=====================

                ! determine region with hightes refinement factor
                if (xref(region) > xref(eddystat(i_stat)%region)) then
                  eddystat(i_stat)%region = region
                  eddystat(i_stat)%is = &
                    int((eddystat(i_stat)%lon-float(xbeg(region)))/dxr + 0.99999) 
                  eddystat(i_stat)%js = &
                    int((eddystat(i_stat)%lat-float(ybeg(region)))/dyr + 0.99999)
                end if
              end if
            end do
          end do
    end if  ! myid = root  
    call reset_eddyflux_accumulator

    return

    ! error handling
1000 call escape_tm( 'read_eddyfluxlist: could not open ' //trim(InputDir)//'/'//trim(eddyfluxlistFFilename))

  end subroutine read_eddyfluxlist


  !=====================================================================
  !=====================================================================

  subroutine init_eddyflux_output
    !=====================================
    ! HDF output for eddyfluxs 
    ! (instantaneous model results)
    !=====================================
    use ParTools ,   only: myid, root
    use Dims,        only : itaui, itaue, ndyn_max
    use file_hdf,    only : Init, WriteAttribute, WriteData, TSds, Done
    use GO,          only : TrcFile, Init, Done, ReadRc
    use global_data, only : rcfile, outdir
    implicit none

    integer                           :: i_stat
    type(TrcFile)                     :: rcF
    character(len=12)                 :: xname
    integer                           :: i, ii, j, iostat, io_status
    type(TSds)                        :: sds_temp


    if (myid /= root) return
    if (write_eddyfluxfiles)then
       call Init( rcF, rcfile,cstatus)
       call ReadRc( rcF, 'output.eddyflux.file', eddyfluxOutFFilename,cstatus)
       call Done( rcF ,cstatus)

       call Init( io_hdf, trim(OutDir)//'/'// trim(eddyfluxOutFFilename),'create' , io_status)
       io_record = 0
       io_nrecords =abs(itaue-itaui)/ndyn_max
       call WriteAttribute( io_hdf, 'io_nrecords', io_nrecords,cstatus)
       call WriteAttribute( io_hdf, 'n_eddystat', n_eddystat,cstatus)
       call WriteAttribute( io_hdf, 'n_spec', eddyflux_nspec,cstatus)
       call WriteAttribute( io_hdf, 'eddyflux_specnames', &
                        names,cstatus)
       call WriteAttribute( io_hdf, 'eddyflux_lat', eddystat(:)%lat,cstatus)
       call WriteAttribute( io_hdf, 'eddyflux_lon', eddystat(:)%lon,cstatus)
       call WriteAttribute( io_hdf, 'eddyflux_height', eddystat(:)%height,cstatus)
       call WriteAttribute( io_hdf, 'eddyflux_index', eddystat(:)%index,cstatus)
       call WriteAttribute( io_hdf, 'eddyflux_region', eddystat(:)%region,cstatus)
       call WriteAttribute( io_hdf, 'eddyflux_is', eddystat(:)%is,cstatus)
       call WriteAttribute( io_hdf, 'eddyflux_js', eddystat(:)%js,cstatus)
       call WriteAttribute( io_hdf, 'eddyflux_names', eddystat(:)%name,cstatus)
       call WriteAttribute( io_hdf, 'eddyflux_ident', eddystat(:)%ident,cstatus)

       allocate(Sds(eddyflux_nspec, n_interpol))
       do i=1, eddyflux_nspec
          do ii= 1,n_interpol
             xname = trim(names(i))//c_interpol(ii)
             call init(Sds(i,ii), io_hdf, xname, (/n_eddystat, io_nrecords/), 'real(4)',cstatus)
             call SetDim( sds(i,ii), 0, 'n_eddyfluxs', 'eddyflux number', (/ (j, j=1,n_eddystat) /) ,cstatus)
             call SetDim( sds(i,ii), 1, 'n_records ', 'time slot #', (/ (j, j=1,io_nrecords) /) ,cstatus)
             call WriteAttribute(Sds(i,ii), 'Unit', 'mole per m2 per second',cstatus)
          enddo
       enddo
       do i=1,n_metdat
          call init(sds_meteo(i), io_hdf, c_metdat(i), (/n_eddystat, io_nrecords/), 'real(4)',cstatus)
          call SetDim( sds_meteo(i), 0, 'n_eddyfluxs', 'eddyflux number', (/ (j, j=1,n_eddystat) /),cstatus)
          call SetDim( sds_meteo(i), 1, 'n_records ', 'time slot #', (/ (j, j=1,io_nrecords) /) ,cstatus)
          call WriteAttribute(Sds_meteo(i), 'Unit', u_metdat(i),cstatus)
       enddo
       call init(sds_idate, io_hdf, 'idate', (/6, io_nrecords/), 'integer(2)',cstatus)
       call SetDim( sds_idate, 1, 'n_records ', 'time slot #', (/ (j, j=1,io_nrecords) /) ,cstatus)
       call WriteAttribute(Sds_idate, 'Unit', 'year month day hour minutes seconds',cstatus)
   end if

  end subroutine init_eddyflux_output

  !=========================================================================
  !=========================================================================


  subroutine close_eddyfluxfiles

    use ParTools , only: myid, root
    use file_hdf,  only: Done
    implicit none
    integer                           :: io_status, i, ii

    if (myid == root) then
       print *, myid, root, 'Write eddyflux files'
       if (write_eddyfluxfiles)then
          do i=1, eddyflux_nspec
             do ii= 1,n_interpol
                call Done(sds(i,ii),io_status)
             enddo
          enddo
          do i=1,n_metdat
             call Done(sds_meteo(i),io_status)
          enddo
          call Done(sds_idate,io_status)
          deallocate(Sds)
          call Done(io_hdf, io_status)
          print *, 'eddyflux file closed with status:', io_status
       endif
    end if   ! myid == root
    call free_eddyfluxsfields
  end subroutine close_eddyfluxfiles


  subroutine output_eddyflux(region)
    !=====================================================================
    !=====================================================================

#ifdef MPI  
    use mpi_const !, only: myid, root
    use mpi_comm, only: gather_tracer_k, gather_tracer_t, stopmpi, gather_tracer_k_3d
#endif
    use ParTools, only: myid, root, root_t, root_k
    use emission_common, only: co2_bio, co2_ff, co2_fires, co2_ocean
    use datetime,    only : tau2date
    use toolbox,     only : escape_tm
    use Meteo,       only : gph_dat, poro=>oro, temper_dat, pu_dat, pv_dat
    use global_data, only : mass_dat,  conv_dat

    implicit none

    ! input/output
    integer,intent(in)  :: region

    ! local

    integer                         :: i_interpol           ! interpolation procedures:   
    ! 1: no interpolation
    ! 2: interpolation based on slopes
    ! 3: 3D linear interpolations


    integer                         :: i,j,l                ! grid cell indices
    integer                         :: is,js,ls             ! i,j,l index of grid cell in which eddyflux lis located
    integer                         :: isn,jsn,lsn          ! i,j,l index of grid cell which is taken as neighbour for linear interpolation
    integer                         :: lst, lstn            ! MPI implementation
    integer                         :: js_north, js_south   ! j index of grid cell for neighbours for windspeed_v interpolation
    integer                         :: is_west              ! i index of grid cell for neighbour for windspeed_u interpolation
    integer                         :: n                    ! index of tracer
    integer                         :: i_stat               ! index of eddyflux
    real                            :: ris,rjs,rls          ! deviation of eddyflux from center of the grid cell (-0.5 ... +0.5)
    real                            :: wcx,wcy,wcz          ! x,y,z-weight of grid cell (1.0 ... 0.5) 
    real                            :: dxr, dyr             ! x,y resolution (in degrees) for current region 

    real,dimension(0:lm(region))    :: height               ! height of vertical grid boundaries 
    integer, dimension(6)           :: idate_f              ! current idate for region 
    integer                         :: sunit                ! unit number for eddyflux output file
    real                            :: eddy_interpol          ! tracer mass, interpolated to eddyflux location 
    integer                         :: ispec                ! index over activated tracers

    !=============
    ! meteo output
    !=============
    real,dimension(n_eddystat,n_metdat) :: metdat               ! meteodata (interpolated) at eddyfluxs
    real,dimension(:,:,:)  ,pointer :: m                    ! air mass
    real,dimension(:,:,:)  ,pointer :: gph                  ! geopotential height
    real,dimension(:,:)    ,pointer :: oro                  ! orography
    real,dimension(:,:,:), pointer  :: T                    ! temperature
    real,dimension(:,:,:), pointer  :: phlb                 ! pressure grid boundaries
    real,dimension(:,:,:), pointer  :: pu                   ! mass flux x-direction [kg/s]
    real,dimension(:,:,:), pointer  :: pv                   ! mass flux y-direction [kg/s]
    real,dimension(:,:), pointer    :: blh                  ! boundary layer height [m]
    real,dimension(:,:), pointer    :: tmflux               ! TM5 flux in mol/m2/s
    real                            :: T_interpol           ! temperature, interpolated to eddyflux location 
    real                            :: p_interpol           ! pressure, interpolated to eddyflux location 
    real                            :: windspeed_u_interpol ! wind speed [m/s] x-direction , interpolated to eddyflux location
    real                            :: windspeed_v_interpol ! wind speed [m/s] y-direction , interpolated to eddyflux location
    real                            :: windspeed_interpol   ! wind speed [m/s]             , interpolated to eddyflux location
    real                            :: winddir_interpol     ! wind direction               , interpolated to eddyflux location 
    real                            :: blh_interpol         ! boundary layer height        , interpolated to eddyflux location 
    real                            :: lat_j                ! latitude of grid cell center js  [degrees]
    real                            :: lat_jn               ! latitude of grid cell center jsn [degrees]
    real                            :: ddx_j                ! x-extension of grid cell js    [m]
    real                            :: ddx_jn               ! x-extension of grid cell jsn   [m]
    real                            :: ddy                  ! y-extension of grid cell js    [m]
    real                            :: vmr                  ! volume mi

    integer   :: communicator,root_id,lmr,imr,jmr

    ! start

    ! assign pointers depending on the paralel status
    ! (over levels after chemistry!)
    imr = im(region) ; jmr = jm(region) ; lmr = lm(region)

    ! assign pointers (METEO)
    m => mass_dat(region)%m_t
    gph => gph_dat(region)%data
    oro => poro(region)%surf
    t => temper_dat(region)%data
    pu => pu_dat(region)%data
    pv => pv_dat(region)%data
    blh => conv_dat(region)%blh
    ! point to pressure paralel over tracers (contains all layers!)
    phlb => mass_dat(region)%phlb_t

    if ( myid == root_t ) then 
       ! perform only on root: fluxes are already gathered here
       ! x and y resolution [degrees] for current region

       dyr = dy/yref(region)
       dxr = dx/xref(region)

       ! idate for current region
       call tau2date(itaur(region),idate_f)


       !====================
       ! loop over eddyfluxs
       !====================

       do i_stat=1, n_eddystat

          IF (eddystat(i_stat)%region == region)THEN

             !WP! Sample fluxes from original 1x1 fields, not model grid

             !avoid edges!
             ris = (eddystat(i_stat)%lon+180) + 0.99999 ! 1x1 grid
             rjs = (eddystat(i_stat)%lat+90) + 0.99999 ! 1x1 grid
             is  = int(ris)   ! i-index of grid cell in which eddyflux is located
             js  = int(rjs)   ! j-index of grid cell in which eddyflux is located
             ! x-deviation from the center of the grid cell  (-0.5...+0.5)
             ris = ris-is-0.5
             ! y-deviation from the center of the grid cell  (-0.5...+0.5)
             rjs = rjs-js-0.5

             !=================================
             !the neighbour for x interpolation
             !=================================
             if ( ris .gt. 0 ) then 
                isn = is+1     
             else
                isn = is-1
             endif

             !=================================
             !the neighbour for y interpolation
             !=================================
             if ( rjs .gt. 0 ) then 
                jsn = js+1     
             else
                jsn = js-1
             endif

             ! x- / y-weighting of grid cell in which eddyflux is located
             wcx = (1.0-abs(ris))    ! 1.0 ... 0.5
             wcy = (1.0-abs(rjs))    ! 1.0 ... 0.5


             !=================================================================
             ! if index of neighbour is exceeding range of region set 
             ! neighbour = current cell (i.e. no interpolation)
             ! in case of cyclic x-boundaries take corresponding cyclic i index
             !=================================================================
             
             if ( jsn == 0) jsn=1
             if ( jsn == jm(region)+1 ) jsn=jm(region)  ! isn-->jsn (wouter Peters)
                ! cyclic x-boundaries
             if ( isn == 0 ) isn=im(region)
             if ( isn == im(region)+1 ) isn=1

             l=0 ! sample surface flux

             select case(l)
             case(0)
                ! force eddyflux to model surface
                ls = 1
                rls = -0.5  
             case default
                ls = l       ! eddyflux's layer
                ! the off-set from the center of the layer (-0.5--->+0.5) 
                ! (interpolation is in (m))
                !rls = (eddystat(i_stat)%height - height(l-1)) / &
                !     (height(l)-height(l-1)) - 0.5
             end select
             eddystat(i_stat)%ls = ls  ! store for output to file

             !=================================
             !the neighbour for z interpolation
             !=================================
             if ( rls .gt. 0 ) then 
                lsn = ls+1     
             else
                lsn = ls-1
             end if
             ! z-weighting of grid cell in which eddyflux is located 
             wcz = (1.0-abs(rls))  !.0 ... 0.5

             !=========================================================
             ! if vertical neighbor is 0 (which does not exist)
             ! take vertical layer with l=2 for EXTRApolation to ground
             !=========================================================

             IF(lsn == 0) THEN
               lsn=2
               wcz=1.0-rls  ! 1.0 ... 1.5
             ENDIF    
             IF(lsn == lmr+1) THEN
                !=========================================================
                ! if vertical neighbor is lmr+1 (which does not exist)
                ! -> no interpolation
                !=========================================================
                lsn=lmr      ! no interpolation 
                wcz=1.0
             ENDIF    

             do ispec=1,eddyflux_nspec
                if(ispec==1) tmflux => co2_ff
                if(ispec==2) tmflux => co2_bio
                if(ispec==3) tmflux => co2_ocean

                lst = ls 
                lstn = lsn 
                !========================================
                ! value of grid box without interpolation
                !========================================
                vmr = tmflux(is,js)
                eddystat(i_stat)%c_mean(ispec,1) = &
                   eddystat(i_stat)%c_mean(ispec,1) + vmr
                eddystat(i_stat)%c_std(ispec,1) = &
                   eddystat(i_stat)%c_std(ispec,1) + vmr*vmr
                !========================================
                ! interpolation using slopes  
                !========================================
                !rm-value is obtained from rm + slopes. 
                !slope = rxm = (rm*dX/dx *deltaX/2)
                vmr = tmflux(is,js) + 2.0*(ris*tmflux(is,js) + &
                     rjs*tmflux(is,js) + &
                     rls*tmflux(is,js) )
                eddystat(i_stat)%c_mean(ispec,2) = &
                   eddystat(i_stat)%c_mean(ispec,2) + vmr
                eddystat(i_stat)%c_std(ispec,2) = &
                   eddystat(i_stat)%c_std(ispec,2) + vmr*vmr

                !========================================
                ! 3D linear interpolation 
                !========================================
                ! this PE handles also the neighbour layer
!PB             if ( lstn >= 1 .and. lstn <= lmr) then
!PBi
                if (lstn <= lmr) then    ! lstn = 0 is already excluded (apply vertical EXTRApolation with upper neighbor; see above)
!PBf
                   vmr = wcx      *wcy      *wcz         *tmflux(is,js)   + & 
                        (1.0-wcx)*wcy      *wcz         *tmflux(isn,js)  + &
                        wcx      *(1.0-wcy)*wcz         *tmflux(is,jsn)  + &
                        (1.0-wcx)*(1.0-wcy)*wcz         *tmflux(isn,jsn) + &
                        wcx      *wcy      *(1.0-wcz)   *tmflux(is,js)  + & 
                        (1.0-wcx)*wcy      *(1.0-wcz)   *tmflux(isn,js) + &
                        wcx      *(1.0-wcy)*(1.0-wcz)   *tmflux(is,jsn) + &
                        (1.0-wcx)*(1.0-wcy)*(1.0-wcz)   *tmflux(isn,jsn) 
                else   !other PE calculates neighbouring contribution
!PB                   rm_interpol = wcx      *wcy      *wcz         *rm(is,js,lst,n)   /m(is,js,lst)   + & 
!PB                        (1.0-wcx)*wcy      *wcz         *rm(isn,js,lst,n)  /m(isn,js,lst)  + &
!PB                        wcx      *(1.0-wcy)*wcz         *rm(is,jsn,lst,n)  /m(is,jsn,lst)  + &
!PB                        (1.0-wcx)*(1.0-wcy)*wcz         *rm(isn,jsn,lst,n) /m(isn,jsn,lst) 
!PB                   do NOT apply factor wcz here (no interpolation in z direction in this case) !!!!!

                   vmr = wcx      *wcy          *tmflux(is,js)   + & 
                                 (1.0-wcx)*wcy          *tmflux(isn,js)  + &
                                 wcx      *(1.0-wcy)    *tmflux(is,jsn)  + &
                                 (1.0-wcx)*(1.0-wcy)    *tmflux(isn,jsn) 

                endif
                eddystat(i_stat)%c_mean(ispec,3) = &
                   eddystat(i_stat)%c_mean(ispec,3) + vmr
                eddystat(i_stat)%c_std(ispec,3) = &
                   eddystat(i_stat)%c_std(ispec,3) + vmr*vmr

                nullify(tmflux)

             end do  !ispec



                !======================
                !meteo data at eddyfluxs
                !======================
             if (output_eddyfluxs_meteo) then              

                 !WP! redo height calculation on model grid to sample model
                 !WP! parameters

                 !avoid edges!
                 ris = (eddystat(i_stat)%lon-float(xbeg(region)))/dxr + 0.99999
                 rjs = (eddystat(i_stat)%lat-float(ybeg(region)))/dyr + 0.99999
                 is  = int(ris)   ! i-index of grid cell in which eddyflux is located
                 js  = int(rjs)   ! j-index of grid cell in which eddyflux is located
                 ! x-deviation from the center of the grid cell  (-0.5...+0.5)
                 ris = ris-is-0.5
                 ! y-deviation from the center of the grid cell  (-0.5...+0.5)
                 rjs = rjs-js-0.5

                 !=================================
                 !the neighbour for x interpolation
                 !=================================
                 if ( ris .gt. 0 ) then 
                    isn = is+1     
                 else
                    isn = is-1
                 endif

                 !=================================
                 !the neighbour for y interpolation
                 !=================================
                 if ( rjs .gt. 0 ) then 
                    jsn = js+1     
                 else
                    jsn = js-1
                 endif

                 ! x- / y-weighting of grid cell in which eddyflux is located
                 wcx = (1.0-abs(ris))    ! 1.0 ... 0.5
                 wcy = (1.0-abs(rjs))    ! 1.0 ... 0.5


                 !=================================================================
                 ! if index of neighbour is exceeding range of region set 
                 ! neighbour = current cell (i.e. no interpolation)
                 ! in case of cyclic x-boundaries take corresponding cyclic i index
                 !=================================================================
                 
                 if ( jsn == 0) jsn=1
                 if ( jsn == jm(region)+1 ) jsn=jm(region)  ! isn-->jsn (wouter Peters)
                 if ( xcyc(region) == 0 ) then
                    ! non-cyclic boundaries
                    if ( isn == 0) isn=1
                    if ( isn == im(region)+1 ) isn=im(region)
                 else
                    ! cyclic x-boundaries
                    if ( isn == 0 ) isn=im(region)
                    if ( isn == im(region)+1 ) isn=1
                 end if

                 !============================================
                 ! interpolate the gph to xy position of eddyflux
                 !============================================
                 !ls = 1   !layer
                 do l=0,lm(region)
                    !CMK note: gph now from 1-->lm+1 (dec 2002)
                    height(l) =     wcx *      wcy*  gph(is,js,l+1)  + &
                               (1.0-wcx)*      wcy*  gph(isn,js,l+1) + &
                                    wcx *(1.0-wcy)*  gph(is,jsn,l+1) + &
                               (1.0-wcx)*(1.0-wcy)*  gph(isn,jsn,l+1)
                 end do

                 !===========================
                 ! determine layer of eddyflux
                 !===========================
                 !do l=0,lm(region)
                 !   if(height(l) .gt. eddystat(i_stat)%height) exit
                 !end do
                 l=0 ! sample surface flux

                 select case(l)
                 case(0)
                    ! force eddyflux to model surface
                    ls = 1
                    rls = -0.5  
                 case default
                    ls = l       ! eddyflux's layer
                    ! the off-set from the center of the layer (-0.5--->+0.5) 
                    ! (interpolation is in (m))
                    rls = (eddystat(i_stat)%height - height(l-1)) / &
                         (height(l)-height(l-1)) - 0.5
                 end select
                 eddystat(i_stat)%ls = ls  ! store for output to file

                 !=================================
                 !the neighbour for z interpolation
                 !=================================
                 if ( rls .gt. 0 ) then 
                    lsn = ls+1     
                 else
                    lsn = ls-1
                 end if
                 ! z-weighting of grid cell in which eddyflux is located 
                 wcz = (1.0-abs(rls))  !.0 ... 0.5

                 !=========================================================
                 ! if vertical neighbor is 0 (which does not exist)
                 ! take vertical layer with l=2 for EXTRApolation to ground
                 !=========================================================

                 IF(lsn == 0) THEN
                   lsn=2
                   wcz=1.0-rls  ! 1.0 ... 1.5
                 ENDIF    
                 IF(lsn == lmr+1) THEN
                    !=========================================================
                    ! if vertical neighbor is lmr+1 (which does not exist)
                    ! -> no interpolation
                    !=========================================================
                    lsn=lmr      ! no interpolation 
                    wcz=1.0
                 ENDIF    

                   !============================================
                   ! interpolate the blh to xy position of eddyflux
                   !============================================
                   blh_interpol =     wcx *      wcy*  blh(is,js)  + &
                                     (1.0-wcx)*      wcy*  blh(isn,js) + &
                                     wcx *(1.0-wcy)*  blh(is,jsn) + &
                                     (1.0-wcx)*(1.0-wcy)*  blh(isn,jsn)
                   eddystat(i_stat)%metdat(7) = &
                      eddystat(i_stat)%metdat(7) + blh_interpol
                   !====================================
                   ! 3D linear interpolation Temperature
                   !====================================
                   T_interpol = &
                        wcx      *wcy      *wcz       *T(is,js,ls)    + & 
                        (1.0-wcx)*wcy      *wcz       *T(isn,js,ls)   + &
                        wcx      *(1.0-wcy)*wcz       *T(is,jsn,ls)   + &
                        (1.0-wcx)*(1.0-wcy)*wcz       *T(isn,jsn,ls)  + &
                        wcx      *wcy      *(1.0-wcz) *T(is,js,lsn)   + & 
                        (1.0-wcx)*wcy      *(1.0-wcz) *T(isn,js,lsn)  + &
                        wcx      *(1.0-wcy)*(1.0-wcz) *T(is,jsn,lsn)  + &
                        (1.0-wcx)*(1.0-wcy)*(1.0-wcz) *T(isn,jsn,lsn) 

                   eddystat(i_stat)%metdat(1) = &
                      eddystat(i_stat)%metdat(1) + T_interpol
                   !====================================
                   ! 3D linear interpolation pressure
                   !====================================
                   p_interpol =(((0.5-rls) * phlb(is,js,ls)  + (0.5+rls) * phlb(is,js,ls+1))  * wcx         * wcy       + &
                        ((0.5-rls) * phlb(isn,js,ls) + (0.5+rls) * phlb(isn,js,ls+1)) * (1.0-wcx)   * wcy       + &
                        ((0.5-rls) * phlb(is,jsn,ls) + (0.5+rls) * phlb(is,jsn,ls+1)) * wcx         * (1.0-wcy) + &
                        ((0.5-rls) * phlb(isn,jsn,ls)+ (0.5+rls) * phlb(isn,jsn,ls+1))* (1.0-wcx)   * (1.0-wcy) ) *0.01 ![Pa] -> [hPa]

                   eddystat(i_stat)%metdat(2) = &
                      eddystat(i_stat)%metdat(2) + p_interpol

                   !====================================
                   ! 3D windspeed_u_interpol (x-direction)
                   !====================================

                   ! pu is eastward flux through east grid box boundary
                   ! windspeed_u_interpol wind component from east     

                   ! latitude of grid cell center js  [degrees]
                   lat_j =ybeg(region)+(js -0.5)*dyr
                   ! latitude of grid cell center jsn [degrees]
                   lat_jn=ybeg(region)+(jsn-0.5)*dyr
                   ! x-extension of grid cell js   [m]
                   ddx_j =ae * twopi * dxr / 360.0 * cos(lat_j*twopi/360.0)
                   ! x-extension of grid cell jsn  [m]
                   ddx_jn=ae * twopi * dxr / 360.0 * cos(lat_jn*twopi/360.0)

                   is_west=is-1
                   if ( xcyc(region) == 0 ) then
                      ! non-cyclic boundaries
                      if ( is_west == 0 ) is_west=1
                   else
                      ! cyclic x-boundaries
                      if ( is_west == 0 ) is_west=im(region)
                   end if

                   !west border                         !east border  
                   windspeed_u_interpol=-(((0.5-ris) * pu(is_west,js,ls)/m(is_west,js,ls)                                   + &
                        (0.5+ris) * pu(is,js,ls)/ m(is,js,ls) )      * &
                        ddx_j  *      wcy *      wcz          + &  
                        ((0.5-ris) * pu(is_west,jsn,ls)/m(is_west,jsn,ls)                                 + & 
                        (0.5+ris) * pu(is,jsn,ls)/m(is,jsn,ls) )     * &
                        ddx_jn * (1.0-wcy)*      wcz          + &  
                        ((0.5-ris) * pu(is_west,js,lsn)/m(is_west,js,lsn)                                 + &
                        (0.5+ris) * pu(is,js,lsn)/ m(is,js,lsn))    * &
                        ddx_j  *      wcy * (1.0-wcz)         + &  
                        ((0.5-ris) * pu(is_west,jsn,lsn)/m(is_west,js,lsn)                                + &
                        (0.5+ris) * pu(is,jsn,lsn)/ m(is,jsn,lsn) ) * &
                        ddx_jn * (1.0-wcy)* (1.0-wcz)         )
                   !====================================
                   ! 3D windspeed_v_interpol (y-direction)
                   !====================================

                   ! pv northward flux through south grid box boundary 
                   ! windspeed_v_interpol wind component from north    

                   ddy =ae * twopi * dyr / 360.0 ! y-extension of grid cell [m]

                   js_south = js-1
                   if (js_south==0) js_south=1

                   js_north = js+1
                   if (js_north==(jm(region)+1)) js_north=jm(region)

                   !south border                  !north border  
                   windspeed_v_interpol=-(( &
                        (0.5-rjs)*pv(is,js,ls)/m(is,js_south,ls)       + &
                        (0.5+rjs)*pv(is,js_north,ls)/ m(is,js,ls) )     * &
                        wcx *      wcz           + &  
                        ((0.5-rjs)*pv(isn,js,ls)/m(isn,js_south,ls)    + &
                        (0.5+rjs)*pv(isn,js_north,ls)/ m(isn,js,ls) )   * &
                        (1.0 -  wcx)*      wcz           + &     
                        ((0.5-rjs)*pv(is,js,lsn)/m(is,js_south,lsn)    + &
                        (0.5+rjs)*pv(is,js_north,lsn)/ m(is,js,lsn) )   * &
                        wcx * (1.0-wcz)          + &  
                        ((0.5-rjs)*pv(isn,js,lsn)/m(isn,js_south,lsn)  + &
                        (0.5+rjs)*pv(isn,js_north,lsn)/ m(isn,js,lsn) ) * &
                        (1.0 -  wcx)* (1.0-wcz)        ) * &
                        ddy 


                   windspeed_interpol = &
                        sqrt(windspeed_u_interpol**2 + windspeed_v_interpol**2)

                   eddystat(i_stat)%metdat(3) = &
                      eddystat(i_stat)%metdat(3) + windspeed_u_interpol
                   eddystat(i_stat)%metdat(4) = &
                      eddystat(i_stat)%metdat(4) + windspeed_v_interpol
                   eddystat(i_stat)%metdat(5) = &
                      eddystat(i_stat)%metdat(5) + windspeed_interpol

                endif  ! meteo out...
                eddystat(i_stat)%ncount = eddystat(i_stat)%ncount + 1
           end if ! in region
       end do ! end i_stat loop

    endif  ! myid = root_t

    nullify(m)
    nullify(gph)
    nullify(oro)
    nullify(t)
    nullify(pu)
    nullify(pv)
    nullify(phlb)
    nullify(blh)

  end subroutine output_eddyflux



  subroutine reset_eddyflux_accumulator

    !==================================================================
    ! initialize accumulation arrays for average eddyflux concentration
    !==================================================================
    use ParTools, only: root_t, myid

    implicit none

    integer i_stat

    if (myid /= root_t) return

    do i_stat=1, n_eddystat
       eddystat(i_stat)%c_mean(:,:) = 0.0
       eddystat(i_stat)%c_std(:,:)  = 0.0
       eddystat(i_stat)%metdat(:)   = 0.0
       eddystat(i_stat)%ncount  = 0
    end do

  end subroutine reset_eddyflux_accumulator
!
!
!
  subroutine evaluate_eddyflux
    use dims, only  : idate
    use ParTools, only: myid, root
    use file_hdf,    only : Init, WriteData, TSds, Done, select
    use file_hdf,    only : CheckInfo
    use toolbox,     only : escape_tm
    !==================================================================
    ! evaluate mean concentration and std
    !==================================================================

    implicit none

    integer i_stat, i, ii, io_status
    integer n_data, isds, nsds, itp
    real temp
    character(len=12)                    :: xname
    real, dimension(:,:), allocatable    :: cout                 ! to get output
    integer, dimension(:,:), allocatable :: iidate
    real                                 :: windspeed_u_interpol ! wind speed [m/s] x-direction , interpolated to eddyflux location
    real                                 :: windspeed_v_interpol ! wind speed [m/s] y-direction , interpolated to eddyflux location
    real                                 :: windspeed_interpol   ! wind speed [m/s]             , interpolated to eddyflux location
    real                                 :: winddir_interpol     ! wind direction               , interpolated to eddyflux location 
    if(myid /= root) return

    do i_stat=1, n_eddystat
       n_data = eddystat(i_stat)%ncount
       if (n_data > 0) then
          eddystat(i_stat)%c_mean(:,:) = eddystat(i_stat)%c_mean(:,:) / n_data
          eddystat(i_stat)%metdat(:)   = eddystat(i_stat)%metdat(:)   / n_data
          if (n_data > 1) then
             do i=1, eddyflux_nspec
                do ii= 1, n_interpol
                   temp = ((eddystat(i_stat)%c_std(i,ii) -  n_data * ((eddystat(i_stat)%c_mean(i,ii))**2) ) / (n_data-1) )
                   if(temp > 0) then 
                      eddystat(i_stat)%c_std(i,ii) = sqrt(temp)
                   else
                      eddystat(i_stat)%c_std(i,ii) = 0.0
                   endif
                enddo
             enddo
          else
             eddystat(i_stat)%c_std(:,:)=-1.0
          end if
       else
          eddystat(i_stat)%c_mean(:,:)= -1.0
          eddystat(i_stat)%c_std(:,:) = -1.0
          eddystat(i_stat)%metdat(:)  = -999.
       end if
       windspeed_u_interpol = eddystat(i_stat)%metdat(3)
       windspeed_v_interpol = eddystat(i_stat)%metdat(4)
       windspeed_interpol   = eddystat(i_stat)%metdat(5)

       !====================================
       ! wind direction
       !====================================

       ! definition of winddirection:          
       ! east     :  90
       ! south    : 180
       ! west     : 270
       ! north    : 360
       if ((windspeed_u_interpol > 0) .and. (windspeed_v_interpol > 0))then              ! wind from north east (0...90)
          winddir_interpol = atan(windspeed_u_interpol / windspeed_v_interpol) * 360.0 / twopi
       else if ((windspeed_u_interpol > 0) .and. (windspeed_v_interpol < 0)) then     ! wind from south east (90...180)
          winddir_interpol = 180.0+atan(windspeed_u_interpol / windspeed_v_interpol) * 360.0 / twopi
       else if ((windspeed_u_interpol > 0) .and. (windspeed_v_interpol == 0)) then    ! wind from east (90)
          winddir_interpol = 90.0
       else if ((windspeed_u_interpol < 0) .and. (windspeed_v_interpol > 0)) then     ! wind from north west (270... 360)
          winddir_interpol = 360.0+atan(windspeed_u_interpol / windspeed_v_interpol) * 360.0 / twopi
       else if ((windspeed_u_interpol < 0) .and. (windspeed_v_interpol < 0)) then     ! wind from south west (180...270)
          winddir_interpol = 180.0+atan(windspeed_u_interpol / windspeed_v_interpol) * 360.0 / twopi
       else if ((windspeed_u_interpol < 0) .and. (windspeed_v_interpol == 0)) then    ! wind from west (270)
          winddir_interpol = 270.0
       else if ((windspeed_u_interpol == 0) .and. (windspeed_v_interpol > 0)) then    ! wind from north (360)
          winddir_interpol = 360.0
       else if ((windspeed_u_interpol == 0) .and. (windspeed_v_interpol < 0)) then    ! wind from south (180)
          winddir_interpol = 180.0
       else if ((windspeed_u_interpol == 0) .and. (windspeed_v_interpol == 0)) then   ! no wind
          winddir_interpol = -1.0
       else
          call escape_tm('output_eddyflux: error wind direction')
       end if
       eddystat(i_stat)%metdat(6) = winddir_interpol
    end do
  end subroutine evaluate_eddyflux

  subroutine write_eddyflux
    use dims, only  : idate
    use ParTools, only: myid, root
    use file_hdf,    only : Init, WriteData, TSds, Done
    use file_hdf,    only : WriteAttribute
    use toolbox,     only : escape_tm
    !==================================================================
    ! write mean concentration
    !==================================================================
    implicit none

    integer i, ii, io_status, i_stat
    integer n_data, isds, nsds, itp
    real, dimension(:,:), allocatable    :: cout                 ! to get output
    integer, dimension(:,:), allocatable :: iidate

    if(myid /= root) return
    ! output
    if (io_record == 0) then
       call WriteAttribute( io_hdf, 'eddyflux_ls', eddystat(:)%ls,cstatus)
    endif

    allocate(cout(n_eddystat,1))

    do i=1, eddyflux_nspec
       do ii= 1,n_interpol
          do i_stat = 1, n_eddystat
            cout(i_stat,1) = eddystat(i_stat)%c_mean(i,ii)
          enddo
          call writedata(sds(i,ii), cout,  cstatus,start=(/0,io_record/))
       enddo
    enddo
    do i=1, n_metdat
       do i_stat = 1, n_eddystat
          cout(i_stat,1) = eddystat(i_stat)%metdat(i)
       enddo
       call writedata(sds_meteo(i), cout, cstatus, start=(/0,io_record/))
    enddo
    deallocate(cout)
    allocate(iidate(6,1))
    iidate(:,1) = idate
    call writedata(sds_idate, iidate, cstatus, start=(/0,io_record/))
    deallocate(iidate)
    io_record = io_record + 1
  end subroutine write_eddyflux


end module user_output_eddyflux
